FreeRefills = LibStub("AceAddon-3.0"):NewAddon("FreeRefills","AceEvent-3.0","AceConsole-3.0")
local L = LibStub("AceLocale-3.0"):GetLocale("FreeRefills")

function FreeRefills:OnInitialize()
	local defaults = {
		profile = {
			overstock = false,
			merchant = true,
			bank = true,
			refills = {},
		}
	}

	self.db = LibStub("AceDB-3.0"):New("FreeRefillsDB", defaults --[[, char default]])

	local options
	LibStub("AceConfigRegistry-3.0"):RegisterOptionsTable("FreeRefills",function()
		if not options then	-- avoid table creation until necessary
			options = {	
				type = 'group',
				name="FreeRefills",	-- displayname in GUIs
				childGroups="tab",
				
				handler = self,
				set="SetDB",
				get="GetDB",
				
				args = {
					cmdlinetools  = {
						type="group",
						name="cmdlinetools",
						hidden = true,
						cmdHidden = false,
						inline = true,
						args = {
							add = {
								order=1,
								name = L["Add"], type = 'input',
							    desc = L["Add an item to the refill list. (Usage: /freerefills add [Item Link] #)"],
								get = false,
							    set = "AddRefillItem",
								usage = L["[Item Link] <number of items>"],
								guiHidden = true
							},
							addstack = {
								order=2,
								name = L["Add Stack"], type = 'input',
							    desc = L["Add an item to the refill list by stack. (Usage: /freerefills addstack [Item Link] #)"],
							    get = false,
								set = "AddRefillItemStack",
								usage = L["[Item Link] <number of stacks>"],
								guiHidden = true
							},
							del = {
								order=3,
								name = L["Remove"], type = 'input',
							    desc = L["Remove an item from the refill list. (Usage: /freerefills del [Item Link])"],
								get = false,
							    set = "RemoveRefillItem",
								usage = L["[Item Link]"],
								guiHidden = true
							},
							list = {
								order=4,
								name = L["List"], type = 'execute',
							    desc = L["List the items in the refill list."],
							    func = "List",
							},
						},
					},
					gui = {
						cmdHidden = true,
						type="description",
						name = "Use the command line to manage the list of items to be refilled. Use \"/freerefills help\"",
					},
					overstock = {
						order=200,
						name = L["Overstock"], type = 'toggle',
						desc = L["Toggle Overstocking.  If an item is sold in stacks and cannot meet your exact requested amount, you can opt to buy slightly more (Overstock)."],
					},
					merchant = {
						order=201,
						name = L["Process Merchant"], type = 'toggle',
						desc = L["Toggle Merchant Looting."],
					},
					bank = {
						order=202,
						name = L["Process Bank"], type = 'toggle',
						desc = L["Toggle Bank Looting."],
					},
					reset = {
						order=299,
						name = L["Clear"], type = 'execute',
					    desc = L["Clear the refill list completely."],
					    func = "ClearRefills",
					},
					profile = LibStub("AceDBOptions-3.0"):GetOptionsTable(self.db),
				}
			}
		end
		
		return options
	end)
	
	self:RegisterChatCommand("freerefills", "OnChatCommand", true)
	self:RegisterChatCommand("fr", "OnChatCommand", true)
end

FreeRefills.onSets={}
function FreeRefills:SetDB(info,v)
	local name = info[#info]
	self.db.profile[name]=v
	if self.onSets[name] then
		self[self.onSets[name]](self,v)
	end
end

function FreeRefills:GetDB(info)
	return self.db.profile[info[#info]]
end

function FreeRefills:OnChatCommand(input)
	LibStub("AceConfigCmd-3.0"):HandleCommand("freerefills", "FreeRefills", input)
end

FreeRefills.onSets["merchant"]="OnSet_merchant"
function FreeRefills:OnSet_merchant(v)
	self[v and "RegisterEvent" or "UnregisterEvent"](self,"MERCHANT_SHOW")
end

FreeRefills.onSets["bank"]="OnSet_bank"
function FreeRefills:OnSet_bank(v)
	self[v and "RegisterEvent" or "UnregisterEvent"](self,"BANKFRAME_OPENED")
end



function FreeRefills:OnEnable()
	for k,funcname in pairs(self.onSets) do
		self[funcname](self, self.db.profile[k])
	end
	
	-- In case we get loaded by AddonLoader... (in which case we don't get to see the events ourself)
	if self.db.profile.bank and select(2,GetContainerNumFreeSlots(BANK_CONTAINER)) then
		self:BANKFRAME_OPENED()
	end
	if self.db.profile.merchant and GetMerchantItemInfo(1) then
		self:MERCHANT_SHOW()
	end
		
end

function FreeRefills:debug(msg)
	--DBG self:Print("|cff0099CC"..msg.."|r")
end

--[[--------------------------------------------------------------------------------
  Merchant Processing
-----------------------------------------------------------------------------------]]

function FreeRefills:MERCHANT_SHOW()
	self:debug("Merchant Opened")
	for itemID, amount in pairs(self.db.profile.refills) do
		local merchSlot = self:ItemInStock(itemID)
		if merchSlot then
			self:debug("Item Found in Merchant Slot #"..merchSlot)
			local amountInBags = GetItemCount(itemID)
			
			if amountInBags >= amount then
				self:debug("Already have enough in bags.")
			else
				local itemName, itemLink, itemRarity, itemLevel, itemMinLevel, itemType, itemSubType, itemStackCount, itemEquipLoc, itemTexture
					= GetItemInfo(itemID)
				local name, texture, price, quantity, numAvailable, isUsable, extendedCost 
					= GetMerchantItemInfo(merchSlot)
				if numAvailable<0 then numAvailable=9e9 end
				
				local toBuy = (amount-amountInBags)/quantity
				if self.db.profile.overstock then
					toBuy = ceil(toBuy)
				else
					toBuy = floor(toBuy)
				end
				
				toBuy=min(toBuy,numAvailable)

				local fullStackQty = floor(itemStackCount / quantity)
				assert(fullStackQty>=1)
				
				while toBuy>0 do
					local now = min(toBuy,fullStackQty)
					self:Print(format("  %2u x %s", now*quantity, itemLink))
					BuyMerchantItem(merchSlot, now)
					toBuy=toBuy-now
				end
			end
		end
	end
end

function FreeRefills:ItemInStock(itemLink)
	self:debug("Searching for Item: "..itemLink)
	local itemName = GetItemInfo(itemLink)
	local numItems = GetMerchantNumItems()
	for i = 1, numItems do
		if GetMerchantItemInfo(i) == itemName then		-- compares names
			return i
		end
	end
end



--[[--------------------------------------------------------------------------------
  Bank Processing
-----------------------------------------------------------------------------------]]

function FreeRefills:BANKFRAME_OPENED()
	self:debug("Bank Opened")
	for itemID,amount in pairs(self.db.profile.refills) do
		local amountInBags = GetItemCount(itemID)
		local amountInBank = GetItemCount(itemID, true) - amountInBags
		if amountInBank > 0 then
			local amountNeeded = amount - amountInBags
			if amountNeeded<0 then
				self:debug("Already have enough in bags.")
			else
				if amountNeeded > amountInBank then
					amountNeeded = amountInBank end
				self:debug(amountNeeded.." item(s) to be looted.")
				self:LootBank(itemID, min(amountNeeded, amountInBank))
			end
		end
	end
end

local function linkDecode(link)
	if link then
		return select(3, link:find("Hitem[^|]+|h%[([^[]+)%]"))
	end
end

function FreeRefills:LootBank(itemID, amountNeeded)
	self:debug("Bank Looting Logic: "..itemID.." - "..amountNeeded)
	
	--Scan the main bank first
	for slot = 1, NUM_BANKGENERIC_SLOTS do
		local slotID = GetContainerItemLink(BANK_CONTAINER, slot)
		local checkitemID = linkDecode(itemID)
		local checkslotID = linkDecode(slotID)
		if checkslotID == checkitemID then
			local slotCount = select(2, GetContainerItemInfo(BANK_CONTAINER, slot))
			self:debug(slotCount.." Item(s) Found in Bank:Slot "..slot)
			if slotCount <= amountNeeded then --We don't have to split this stack, we can cheat and use UseContainerItem
				self:debug("Looting Entire Stack")
				UseContainerItem(BANK_CONTAINER, slot)
				amountNeeded = amountNeeded - slotCount
				self:debug(slotCount.." Looted - Need "..amountNeeded.." more item(s)")
			else
				SplitContainerItem(BANK_CONTAINER, slot, amountNeeded)
				for bag = NUM_BAG_SLOTS, 0, -1 do
					self:debug("Scanning Bag #"..bag)

					if select(7, GetItemInfo(GetBagName(bag))) == "Bag" then
						local size = GetContainerNumSlots(bag)
						for slot = 1, size, 1 do
							local slotID = GetContainerItemLink(bag,slot)
							if not slotID then
								if bag == 0 then
									PutItemInBackpack()
								else
									PutItemInBag(bag + 19)
								end
							end
						end
					end
				end
				local amountLooted = amountNeeded
				amountNeeded = amountNeeded - amountLooted -- This should always result in 0
				self:debug(amountLooted.." Looted - Need "..amountNeeded.." more item(s)")
			end
		end
		if amountNeeded == 0 then self:debug(amountNeeded.." item(s) # "..itemID.." needed.  Next item.") return end
	end
	self:debug("Main Bank didn't cover it, scanning additional Bags")
	--Now Scan bank bags
	for bag = NUM_BAG_SLOTS+1, NUM_BAG_SLOTS+NUM_BANKBAGSLOTS, 1 do
		local size = GetContainerNumSlots(bag)
		if (size > 0) then
			for slot = 1, size do
				local slotID = GetContainerItemLink(bag, slot)
				local checkitemID = linkDecode(itemID)
				local checkslotID = linkDecode(slotID)
				if checkslotID == checkitemID then
					local slotCount = select(2, GetContainerItemInfo(bag, slot))
					self:debug(slotCount.." Item(s) Found in Bag "..bag..":Slot "..slot..". Need to loot "..amountNeeded)
					if slotCount <= amountNeeded then --We don't have to split this stack, we can cheat and use UseContainerItem
						self:debug("Looting Entire Stack")
						UseContainerItem(bag, slot)
						amountNeeded = amountNeeded - slotCount
						self:debug(slotCount.." Looted - Need "..amountNeeded.." more item(s)")
					else
						SplitContainerItem(bag, slot, amountNeeded)
						for bag = NUM_BAG_SLOTS, 0, -1 do
							self:debug("Scanning Bag #"..bag)
							local bagName = GetBagName(bag)
							if bagName:find("Quiver", 1) == nil and bagName:find("Pouch", 1) == nil then
								local size = GetContainerNumSlots(bag)
								for slot = 1, size, 1 do
									local slotID = GetContainerItemLink(bag,slot)
									if not slotID then
										if bag == 0 then
											self:debug("Item in Bag#"..bag)
											PutItemInBackpack()
										else
											self:debug("Item in Bag#"..bag)
											PutItemInBag(bag + 19)
										end
									end
								end
							end
						end
						local amountLooted = amountNeeded
						amountNeeded = amountNeeded - amountLooted -- This should always result in 0
						self:debug(amountLooted.." Looted - Need "..amountNeeded.." more item(s)")
					end
				end
				if amountNeeded == 0 then self:debug(amountNeeded.." item(s) # "..itemID.." needed.  Next item.") return end
			end
		end
	end
end

--[[--------------------------------------------------------------------------------
  Command Handlers
-----------------------------------------------------------------------------------]]

function FreeRefills:AddRefillItem(info, msg)
	--add an item to the refill list
	local link, num = strmatch(msg,"(.+)%s(%d+)")
	if not link or not num then
		self:Print(L["Invalid input; input must be '<item> #'."])
		return
	end
	num = tonumber(num)
	self.db.profile.refills[link] = num
	self:Print(L["Added %s - Amount:%d"]:format(link, num))
end

local function GetStackSize(item)
	local itemStack = select(8, GetItemInfo(item))
	return itemStack
end

function FreeRefills:AddRefillItemStack(info, msg)
	--add an item to the refill list by stack
	local link, num = strmatch(msg,"(.+)%s(%d+)")
	if not link or not num then
		self:Print(L["Invalid input; input must be '<item> #'."])
		return
	end
	num = tonumber(num)
	num = num * GetStackSize(link)
	self.db.profile.refills[link] = num
	self:Print(L["Added %s - Amount:%d"]:format(link, num))
end

function FreeRefills:RemoveRefillItem(info, msg)
	--remove an item from the refill list
	if not self.db.profile.refills[msg] then return end
	self.db.profile.refills[msg] = nil
	self:Print(L["Removed %s."]:format(msg))
end

function FreeRefills:ClearRefills()
	--clear the refill list
	self.db.profile.refills = {}
	self:Print(L["Item List Cleared"])
end

function FreeRefills:List()
	--report back which items are currently on the refill list, and their quanitity
	self:Print(L["Refill List: "])
	local notempty
	for itemID,amount in pairs(self.db.profile.refills) do
		self:Print("     "..itemID.." ("..amount..")")
		notempty = 1
	end
	if not notempty then self:Print(L["     None"]) end
end


